# encoding: utf-8

"""Access Control List Security Extension

Predicate-based security for your applications, extensions, and reusable components.

* This extension is **available in**: `WebCore>=2.0.3,<2.1.0`
* This extension **requires [Python 3](https://www.python.org/)**.
* This extension has **no external package dependencies**.
* This extension is **available with the name** `acl` in the `web.ext` namespace.
* This extension adds the following **context attributes**:
	* `acl`
* This extension uses **namespaced plugins** from:
	* `web.acl.predicate`

1. [**Introduction**](#introduction)
	2. [Operation](#operation)
	3. [Predicates](#predicates)
2. [**Usage**](#usage)
	1. [Enabling the Extension](#enabling-the-extension)
		1. [_Imperative Configuration_](#imperative-configuration)
		2. [_Declarative Configuration_](#declarative-configuration)
	2. [Defining ACLs](#defining-acls)
		1. [_Explicit Usage_](#explicit-usage)
		2. [_Decoration_](#decoration)
		3. [_Endpoint Return Values_](#endpoint-return-values)


# Introduction

This extension provides a method of collecting rules from objects during dispatch descent and subsequently evaluating
them. This serves the purpose of an access cotrol list (ACL) by allowing these rules to grant access (return `True`),
explicitly deny access (return `False`), or abstain (return `None`). Additionally, values returned from endpoints will
have the value of their `__acl__` attribute evaluated, if present.

Also provided are a stock set of _predicates_ which allow for basic boolean logic, various nesting patterns, and
provide building blocks for more complex behaviour. It is preferred to access these via the `where` helper object,
whose attributes (also provided as a mapping) are the names of `entry_points` registered plugins.


## Operation

On any endpoint or object leading to an endpoint during _dispatch_, define an `__acl__` attribute or property which
provides an iterable (`set`, `list`, `tuple`, generator, etc.) of _predicate_ objects. Objects may also optionally
specify an `__acl_inherit__` attribute or property, which, if _falsy_, will clear the ACL that had been built so far
for the request.

After a final endpoint has been reached, these rules are evaluated in turn (using the `First` predicate), passing the
request context as their first argument.  Each is called until one either returns `True` to indicate permission has
been explicitly granted, or returns `False` to indicate permission has been explicitly denied. If no predicates decide
to have an opinion, the default action is configurable.


## Predicates

A _predicate_ is any callable object that optionally accepts a context as its first positional parameter. One might
look like:

```python
def always(context=None):
	return True
```

That's it, really. The provided built-in ones are class-based, but the process is the same even if the method has a
strange name like `__call__`.


# Usage

## Enabling the Extension

Before utilizing access control list functionality in your own application you must first enable the extension.

Regardless of configuration route chosen rules may be specified as strings, classes, or callables. Strings are
resolved using the `web.acl.predicate` `entry_point` namespace and further processed. Classes (either directly, or
loaded by plugin name) are instantiated without arguments and their instance used.


### Imperative Configuration

Applications using code to explicitly construct the WebCore `Application` object, but with no particular custom base
ruleset needed, can pass the extension by name. It will be loaded using its `entry_points` plugin reference.

```python
app = Application("Hi.", extensions=['acl'])
```

Applications with a more strict configuration may wish to specify a default rule of `never`. Import the extension
yourself, and specify a default rule.

```python
from web.ext.acl import ACLExtension, when

app = Application("Hi.", extensions=[ACLExtension(default=when.never)])
```

More complex arrangements can be had by specifying rules positionally (their order will be preserved) or by passing a
`policy` iterable by name. These may be combined with the `default` named argument, with them being combined as
`positional` + `policy` + `default`.


### Declarative Configuration

Using a JSON or YAML-based configuration, you would define your application's `extensions` section either with the
bare extension declared:

```yaml
application:
	root: "Hi."
	
	extensions:
		acl:
```

Or, specify a default policy by treating the `acl` entry as an array:

```yaml
application:
	root: "Hi."
	
	extensions:
		acl:
			- never
```

By specifying a singular `default` explicitly:

```yaml
application:
	root: "Hi."
	
	extensions:
		acl:
			default: never
```

Or, finally, by specifying the `policy`, which must be an array, explicitly:

```yaml
application:
	root: "Hi."
	
	extensions:
		acl:
			policy:
				- never
```

Use of `policy` and `default` may be combined, with the default appended to the given policy.


## Defining ACLs

Note: We'll be using [object dispatch](https://github.com/marrow/web.dispatch.object/) for these examples.

First, you're going to need to `from web.ext.acl import when` to get easy access to predicates.


### Explicit Usage

Define an iterable of predicates using the `__acl__` attribute.

```python
class PermissiveController:
	__acl__ = [when.always]
	
	def __init__(self, context):
		pass
	
	def __call__(self):
		return "Hi."
```

For intermediary nodes in descent and return values, such as a "root controller with method" arrangement, you can
define an `__acl__` attribute. The contents of this attribute is collected during processing of dispatch events.


### Decoration

Using the `when` utility as a decorator or decorator generator.

```python
@when(when.never)
class SecureController:
	def __init__(self, context):
		pass
	
	@when(when.always, inherit=False)
	def insecure_resource(self):
		return "Yo."
	
	def __call__(self):
		return "Hi."
```

You can use the `when` predicate accessor as a decorator, defining the predicates for an object as positional
parameters.  The result of calling `when` can be saved used later as a decorator by itself, or as a filter to set that
attribute on other objects.


### Endpoint Return Values

Controlling access to information, not just endpoints.

```python
class Thing:
	__acl__ = [when.never]


def endpoint(context):
	return Thing()
```

In this example, `Thing` will not allow itself to be returned by an endpoint. This process is not recursive.


# Extending

Defining new predicates is fairly straightforward given the very simple interface. However, because many demands of
predicates can be resolved entirely by comparison against a value from the request context, two predicate factories
are provided. These can be used on-demand, or the result can be saved for repeated use later.


### Context Value Matches

Grant or deny access based on a value from the context matching one of several possible values.

```python
deny_console = when.matches(False, 'request.client_addr', None)
local = when.matches(True, 'request.remote_addr', '127.0.0.1', '::1')

@when(deny_console, local, when.matches(True, 'user.admin', True))
def endpoint(context):
	return "Hi."
```

This will grant access to local users and those with the `user.admin` flag equal to `True`, as retrieved from the
context. The `local` predicate has been saved for subsequent use, and demonstrates comparing against any of a number
of allowable values. The first argument is the intended predicate result if a match is made, the second is the value
to traverse and compare, and any remaining arguments are treated as acceptable values.


### Context Value Contains

Grant or deny access based on a value from the context containing one of several possible values.

```python
role = when.contains.partial(True, 'user.role')

@when(role('admin', 'editor'))
def endpoint(context):
	return "Hi."
```

This allows you to easily compare against containers such as lists and sets. Also demonstrataed is the ability to
"partially apply" a predicate, that is, apply some arguments, then apply the rest later.
"""

# ## Imports

from __future__ import unicode_literals

from weakref import proxy
from functools import partial
from itertools import chain
from inspect import isclass
from webob.exc import HTTPForbidden
from marrow.package.host import PluginManager
from marrow.package.loader import traverse

from web.core.util import safe_name
from web.core.compat import Path


# ## Module Globals

log = __import__('logging').getLogger(__name__)

if __debug__:  # Documentation helpers.
	__doc_groups__ = {  # Map collapsable sections.
				'Imperative Configuration': {'config', 'choice'},
				'Declarative Configuration': {'config', 'choice'},
				'Explicit Usage': {'sample', 'captioned'},
				'Decoration': {'sample', 'captioned'},
				'Endpoint Return Values': {'sample', 'captioned'},
				'Context Value Matches': {'sample', 'captioned'},
				'Context Value Contains': {'sample', 'captioned'},
			}


# ## Plugin Manager and Decorator

class _When(PluginManager):
	"""A derivative of a PluginManager that acts as a decorator.
	
	Will assign the `__acl__` property as the tuple of arguments passed.
	"""
	def __call__(self, *acl, **kw):
		inherit = kw.pop('inherit', True)
		
		if __debug__:
			if kw:  # This is the only keyword argument we accept.
				raise TypeError("Unknown keyword arguments: " + ", ".join(sorted(kw)))
		
		def acl_when_inner(target):
			if acl:
				target.__acl__ = acl
			
			if not inherit:
				target.__acl_inherit__ = False
			
			return target
		
		return acl_when_inner

when = _When('web.acl.predicate')  # Easy reference by short name, e.g. when.match(...)


# ## ACL Evaluation Result and ACL List Abstractions

class ACLResult(object):
	__slots__ = ('result', 'predicate', 'path', 'source')
	
	def __init__(self, result, predicate, path=None, source=None):
		self.result = result
		self.predicate = predicate
		self.path = path
		self.source = source
	
	def __bool__(self):
		return bool(self.result)
	
	__nonzero__ = __bool__


class ACL(list):
	def __init__(self, *rules, **kw): # Python 3: , context=None, policy=None):
		super(ACL, self).__init__((None, rule, None) for rule in rules)
		
		context = kw.pop('context', None)
		policy = kw.pop('policy', None)
		
		if __debug__:
			if kw:  # This is the only keyword argument we accept.
				raise TypeError("Unknown keyword arguments: " + ", ".join(sorted(kw)))
		
		self.context = proxy(context) if context else None
		self.policy = policy or ()
	
	@property
	def is_authorized(self):
		for path, predicate, source in self:
			result = predicate() if self.context is None else predicate(self.context)
			
			if __debug__:
				log.debug(repr(predicate) + " (from " + repr(source) + ") voted " + repr(result))
			
			if result is None:
				continue
			
			return ACLResult(result, predicate, path, source)
		
		return ACLResult(None, None, None, None)
	
	def __bool__(self):
		return bool(len(self) or len(self.policy))
	
	__nonzero__ = __bool__
	
	def __iter__(self):
		return chain(super(ACL, self).__iter__(), ((None, i, None) for i in self.policy))
	
	def __repr__(self):
		return '[' + ', '.join(repr(i) for i in self) + ']'


# ## Simple Predicates

class Predicate(object):
	__slots__ = ()
	
	def __call__(self, context=None):
		raise NotImplementedError()
	
	@classmethod
	def partial(cls, *args, **kw):
		"""Retrieve a partially applied constructor for this predicate."""
		return partial(cls, *args, **kw)


class Not(Predicate):
	"""Invert the meaning of another predicate."""
	
	__slots__ = ('predicate', )
	
	def __init__(self, predicate):
		self.predicate = predicate
	
	def __call__(self, context=None):
		result = self.predicate(context) if context else self.predicate()
		
		if result is None:
			return
		
		return not result


class Always(Predicate):
	"""Always grant access."""
	
	__slots__ = ()
	
	def __call__(self, context=None):
		return True

always = Always()  # Convienent singleton to use.


class Never(Predicate):
	"""Always deny access."""
	
	__slots__ = ()
	
	def __call__(self, context=None):
		return False

never = Never()  # Convienent singleton to use.


class First(Predicate):
	"""Authorizes or denies an action on the first non-veto predicate."""
	
	__slots__ = ('predicates', )
	
	def __init__(self, *predicates):
		self.predicates = predicates
	
	def __call__(self, context=None):
		for predicate in self.predicates:
			result = predicate(context) if context else predicate()
			
			if result is None:  # Abstain
				continue
			
			return bool(result)


class All(Predicate):
	"""Authorizes an action only if all predicates authorize the action.
	
	Returns `False` on first failure, `True` if all voting predicates returned `True`, `None` otherwise.
	"""
	
	__slots__ = ('predicates', )
	
	def __init__(self, *predicates):
		self.predicates = predicates
	
	def __call__(self, context=None):
		if context:
			results = (predicate(context) for predicate in self.predicates)
		else:
			results = (predicate() for predicate in self.predicates)
		
		vote = None
		
		for result in results:
			if result is None:  # Abstain
				continue
			
			if not bool(result):  # Exit Early
				return False
			
			vote = True
		
		return vote


class Any(Predicate):
	"""Authorize an action if any predicate authorizes the action.
	
	Returns `True` on first success, `False` if all voting predicates returned `False`, `None` otherwise.
	"""
	
	__slots__ = ('predicates', )
	
	def __init__(self, *predicates):
		self.predicates = predicates
	
	def __call__(self, context=None):
		if context:
			results = (predicate(context) for predicate in self.predicates)
		else:
			results = (predicate() for predicate in self.predicates)
		
		vote = None
		
		for result in results:
			if result is None:  # Abstain
				continue
			
			if bool(result):
				return True
			
			vote = False
		
		return vote


class ContextMatch(Predicate):
	"""Match a variable from the context to one of a set of candidate values.
	
	As per most non-base predicates, this accepts a `grant` value determining if a match should be considered success
	(`True`) or not (`False`), then the attribute to attempt to look up, which may be deep or contain array
	references (see the `traverse` function from the `marrow.package.loader` package), and one or more values to
	consider as a match. In the event the attribute can not be loaded, the `default` (which must be passed as a
	keyword argument) will be used, `None` by default.
	
	Examples might include:
	
	admin = ContextMatch(True, 'user.admin', True)
	local = ContextMatch(True, 'request.remote_addr', '127.0.0.1', '::1')
	"""
	
	__slots__ = ('grant', 'attribute', 'values', 'default')
	SENTINEL = object()  # A singleton used internally for comparison.
	
	def __init__(self, grant, attribute, *values, **kw):
		default = kw.pop('default', None)
		
		if __debug__:
			if kw:  # This is the only keyword argument we accept.
				raise TypeError("Unknown keyword arguments: " + ", ".join(sorted(kw)))
			
			if not values:
				raise TypeError("You must supply one or more values to compare against.")
			
			if grant not in (True, False):
				raise ValueError("The `grant` argument must be `True` (allow) or `False` (deny).")
			
			if default not in (None, True, False):
				raise ValueError("The default may either be `True` (allow), `False` (deny), or `None` (abstain).")
		
		self.grant = grant  # True if we grant access, False if we deny access.
		self.attribute = attribute  # The attribute to retrieve, i.e. "user.admin"
		self.values = values
		self.default = default
	
	def __call__(self, context):
		value = traverse(context, self.attribute, self.SENTINEL)  # Retrieve the value.
		
		if value is self.SENTINEL:
			return self.default
		
		result = any(i == value for i in self.values)  # We do this rather than "in" to support equality comparison.
		
		return self.grant if result else None


class ContextContains(ContextMatch):
	"""Match a variable from the context containing one or more values.
	
	Similar to ContextMatch, except matches the values being "in" the target variable rather than comparing equality.
	
	Examples might include:
	
	reviewer = ContextContains(True, 'user.role', 'reviewer')
	"""
	
	__slots__ = ('grant', 'attribute', 'values', 'default')
	
	def __call__(self, context):
		value = traverse(context, self.attribute, self.SENTINEL)  # Retrieve the value.
		
		if value is self.SENTINEL:
			return self.default
		
		result = any(i in value for i in self.values)
		
		return self.grant if result else None


# ## Extension

class ACLExtension(object):
	"""Access control list extension.
	
	Predicates are gathered as dispatch descends through objects, collecting them from the `__acl__` attribute.
	If any object defines `__acl_inherit__` as a falsy value the ACL gathered so far is cleared before continuing.
	"""
	
	provides = {'acl'}
	extensions = {'web.acl.predicate'}
	context = {'acl'}
	
	def __init__(self, *_policy, **kw): # Python 3: , default=None, policy=None):
		"""Configure the ACL extension by defining an optional base policy.
		
		This policy will be used as the base for every request; these are evaluated first, always.
		"""
		
		default = kw.pop('default', None)
		policy = kw.pop('policy', None)
		
		if __debug__:
			if kw:  # This is the only keyword argument we accept.
				raise TypeError("Unknown keyword arguments: " + ", ".join(sorted(kw)))
		
		if policy is None:
			policy = []
		
		if _policy:
			policy = chain(_policy, policy)
		
		policy = (when[rule] if isinstance(rule, str) else rule for rule in policy)
		policy = (rule() if isclass(rule) else rule for rule in policy)
		
		policy = self.policy = list(policy)
		
		if default:
			policy.append(default)
		
		log.info("ACL extension prepared with defualt policy: " + repr(policy))
	
	# ### Request-Level Callbacks
	
	def prepare(self, context):
		"""Called to prepare the request context by adding an `acl` attribute."""
		
		if __debug__:
			log.debug("Preparing request context with ACL.", extra=dict(request=id(context)))
		
		context.acl = ACL(context=context, policy=self.policy)
	
	def dispatch(self, context, consumed, handler, is_endpoint):
		"""Called as dispatch descends into a tier.
		
		The ACL extension uses this to build up the current request's ACL.
		"""
		
		acl = getattr(handler, '__acl__', ())
		inherit = getattr(handler, '__acl_inherit__', True)
		
		if __debug__:
			log.debug("Handling dispatch event: " + repr(handler) + " " + repr(acl), extra=dict(
					request = id(context),
					consumed = consumed,
					handler = safe_name(handler),
					endpoint = is_endpoint,
					acl = [repr(i) for i in acl],
					inherit = inherit,
				))
		
		if not inherit:
			if __debug__:
				log.debug("Clearing collected access control list.")
			
			del context.acl[:]
		
		context.acl.extend((Path(context.request.path), i, handler) for i in acl)
	
	def mutate(self, context, handler, args, kw):
		if not context.acl:
			if __debug__:
				log.debug("Skipping validation of empty ACL.", extra=dict(request=id(context)))
			
			return
		
		grant = context.acl.is_authorized
		
		if grant.result is False:
			log.error("Endpoint authorization failed: " + repr(grant), extra=dict(
					grant = False,
					predicate = repr(grant.predicate) if grant.predicate else None,
					path = str(grant.path) if grant.path else None,
					source = safe_name(grant.source) if grant.source else None
				))
			raise HTTPForbidden()
		
		elif __debug__:
			log.debug("Successful endpoint authorization: " + repr(grant), extra=dict(
					grant = False,
					predicate = repr(grant.predicate) if grant.predicate else None,
					path = str(grant.path) if grant.path else None,
					source = safe_name(grant.source) if grant.source else None
				))
	
	def transform(self, context, handler, result):
		try:
			acl = result.__acl__
		except AttributeError:
			return result
		
		acl = ACL(*acl, context=context)
		valid = acl.is_authorized
		
		if valid.result is False:
			log.error("Response rejected due to return value authorization failure.", extra=dict(
					grant = False,
					predicate = repr(valid.predicate) if valid.predicate else None,
					path = str(valid.path) if valid.path else None,
					result = safe_name(type(result))
				))
			return HTTPForbidden()
		
		elif __debug__:
			log.debug("Successful response authorization.", extra=dict(
					grant = valid.result,
					predicate = repr(valid.predicate) if valid.predicate else None,
					path = str(valid.path) if valid.path else None,
					result = safe_name(type(result))
				))
		
		return result




# encoding: utf-8

"""The WebCore view registry.

WebCore uses a registry of callables to transform values returned by controllers for use as a response. This
translation process is used to support the built-in types (see the `base` extension) and can be extended by your
own application to support additional types.  This is effectively the "view" component of Model-View-Controller.

A view handler takes the result of calling a controller endpoint and applies the result to the response. It can be
as simple as a bare function that accepts the current `RequestContext` instance plus the value returned by the
endpoint, and either returns a truthy value to indicate successful handling, or a falsy value to indicate that this
view handler could not handle the result. The simplest case is a handler that always "passes the buck" and handles
nothing:

	def ignore(context, result):
		pass

Slightly more useful would be to assign the result directly to the response:

	def stringy(context, result):
		context.response.text = result
		return True

As an example pulled from the "template" extension, you can use an "exit early" strategy to selectively handle
subsets of the type your view is registered against, such as only handling 2-tuples with a specific first value:

	def json(context, result):
		from json import dumps
		if len(result) != 2 or result[0] != 'json':
			return
		context.response.text = dumps(result)
		return True

A view may go so far as to wholly replace the `context.response` object; any callable WSGI application can be
utilized as such. Cooperative response construction is _strongly_ preferred, however.

If you frequently hand back serialized data, you may be able to simplify your controller code and reduce
boilerplate by simply returning your model instances.  By registering a view handler for your model class you can
implement the serialization in a single, centralized location, making security and testing much easier.

When a controller raises an `HTTPError` subclass it is treated as the return value.  As such you can take specific
action on any given error by registering a view handler for the specific exception subclass (i.e. `HTTPNotFound`),
such as rendering a pretty error page.  By default these exceptions are treated as a WSGI application and are
directly used as the response, but only if no more specific handlers are registered.
"""

# ## Imports

from __future__ import unicode_literals

from webob.multidict import MultiDict
from marrow.package.canonical import name
from marrow.package.host import PluginManager

from .compat import py3, pypy


# ## Module Globals

# A standard logger object.
log = __import__('logging').getLogger(__name__)


# ## View Registry

class WebViews(PluginManager):
	"""A `PluginManager` subclass to manage and search plugin and manually-registered views.
	
	This extends plugin naming to support multiple candidates per name, overrides manual registration to log some
	useful messages, and makes instances callable. Executing a `WebViews` instance passing in an object will
	produce a generator yielding candidate views registered to handle that type of object.
	"""
	
	__isabstractmethod__ = False  # Work around a Python 3.4+ issue, since our instances are callable.
	
	# ### Python Protocols
	
	def __init__(self, ctx):
		"""View registry constructor.
		
		The view registry is not meant to be instantiated by third-party software. Instead, access the registry as
		an attribute of the the current `ApplicationContext` or `RequestContext`: `context.view`
		"""
		super(WebViews, self).__init__('web.view')
		self.__dict__['_map'] = MultiDict()
	
	def __repr__(self):
		"""Programmers' representation for development-time diagnostics."""
		return "WebViews({})".format(len(self._map))
	
	def __call__(self, result):
		"""Identify view to use based on the type of result when our instance is called as a function.
		
		This generates a stream of candidates which should be called in turn until one returns a truthy value.
		"""
		
		rtype = type(result)
		
		# Short-circuit on exact matches.
		for candidate in self._map.getall(rtype):
			yield candidate
		
		# More exhaustive search for potentially more crafty use such as ABC, zope.interface, marrow.interface, etc.
		for kind, candidate in self._map.iteritems():
			if kind is rtype: continue  # We've already seen these.
			
			if isinstance(result, kind):
				yield candidate
	
	# ### Plugin Registration
	
	def register(self, kind, handler):
		"""Register a handler for a given type, class, interface, or abstract base class.
		
		View registration should happen within the `start` callback of an extension.  For example, to register the
		previous `json` view example:
		
			class JSONExtension:
				def start(self, context):
					context.view.register(tuple, json)
		
		The approach of explicitly referencing a view handler isn't very easy to override without also replacing the
		extension originally adding it, however there is another approach. Using named handlers registered as discrete
		plugins (via the `entry_point` argument in `setup.py`) allows the extension to easily ask "what's my handler?"
		
			class JSONExtension:
				def start(self, context):
					context.view.register(
							tuple,
							context.view.json
						)
		
		Otherwise unknown attributes of the view registry will attempt to look up a handler plugin by that name.
		"""
		if __debug__:  # In production this logging is completely skipped, regardless of logging level.
			if py3 and not pypy:  # Where possible, we shorten things to just the cannonical name.
				log.debug("Registering view handler.", extra=dict(type=name(kind), handler=name(handler)))
			else:  # Canonical name lookup is not entirely reliable on some combinations.
				log.debug("Registering view handler.", extra=dict(type=repr(kind), handler=repr(handler)))
		
		# Add the handler to the pool of candidates. This adds to a list instead of replacing the "dictionary item".
		self._map.add(kind, handler)
		
		return handler




# encoding: utf-8

"""A `MutableMapping` subclass for use as a request-local context object."""

# ## Imports

from __future__ import unicode_literals

from collections import MutableMapping


# ## Mapping Class

class Context(MutableMapping):
	"""An attribute access dictionary, of a kind.
	
	This utility class is used to cooperatively construct the ApplicationContext (and subsequent RequestContext)
	from the contributions of multiple extensions. The concept of "promotion to a class" is needed in order to enable
	the use of descriptor protocol attributes; without promotion the protocol would not be utilized.
	"""
	
	# M-Morty! We're, *belch*, gonna have to go in deep, Morty!  Elbow deep!
	def _promote(self, name, instantiate=True):
		"""Create a new subclass of Context which incorporates instance attributes and new descriptors.
		
		This promotes an instance and its instance attributes up to being a class with class attributes, then
		returns an instance of that class.
		"""
		
		metaclass = type(self.__class__)
		contents = self.__dict__.copy()
		cls = metaclass(str(name), (self.__class__, ), contents)
		
		if instantiate:
			return cls()
		
		return cls
	
	def __init__(self, **kw):
		"""Construct a new Context instance.
		
		All keyword arguments are applied to the instance as attributes through direct assignment to `__dict__`.
		"""
		self.__dict__.update(kw)
		super(Context, self).__init__()
	
	def __len__(self):
		"""Get a list of the public data attributes."""
		return len([i for i in (set(dir(self)) - self._STANDARD_ATTRS) if i[0] != '_'])
	
	def __iter__(self):
		"""Iterate all valid (public) attributes/keys."""
		return (i for i in (set(dir(self)) - self._STANDARD_ATTRS) if i[0] != '_')
	
	def __getitem__(self, name):
		"""Retrieve an attribute through dictionary access."""
		try:
			return getattr(self, name)
		except AttributeError:
			pass
		
		# We do this here to avoid Python 3's nested exception support.
		raise KeyError(name)
	
	def __setitem__(self, name, value):
		"""Assign an attribute through dictionary access."""
		setattr(self, name, value)
	
	def __delitem__(self, name):
		"""Delete an attribute through dictionary access."""
		try:
			return delattr(self, name)
		except AttributeError:
			pass
		
		# We do this here to avoid Python 3's nested exception support.
		raise KeyError(name)

# We generally want to exclude "default object attributes" from the context's list of attributes.
# This auto-detects the basic set of them for exclusion from iteration in the above methods.
Context._STANDARD_ATTRS = set(dir(Context()))



class ContextGroup(Context):
	"""A managed group of related context additions.
	
	This proxies most attribute access through to the "default" group member.
	
	Because of the possibility of conflicts, all attributes are accessible through dict-like subscripting.
	
	Register new group members through dict-like subscript assignment as attribute assignment is passed through to the
	default handler if assigned.
	"""
	
	default = None
	
	def __init__(self, default=None, **kw):
		if default is not None:
			self.default = default
			default.__name__ = 'default'
		
		for name in kw:
			kw[name].__name__ = name
			self.__dict__[name] = kw[name]
	
	def __repr__(self): 
		return "{0.__class__.__name__}({1})".format(self, ', '.join(sorted(self)))
	
	def __len__(self):
		return len(self.__dict__)
	
	def __iter__(self):
		return iter(set(dir(self)) - self._STANDARD_ATTRS)
	
	def __getitem__(self, name):
		try:
			return getattr(self, name)
		except AttributeError:
			pass
		
		raise KeyError()
	
	def __setitem__(self, name, value):
		self.__dict__[name] = value
	
	def __delitem__(self, name):
		del self.__dict__[name]
	
	def __getattr__(self, name):
		if self.default is None:
			raise AttributeError()
		
		return getattr(self.default, name)
	
	def __setattr__(self, name, value):
		if self.default is not None:
			return setattr(self.default, name, value)
		
		self.__dict__[name] = value
	
	def __delattr__(self, name):
		if self.default is not None:
			return delattr(self.default, name)
		
		self.__dict__[name] = None
		del self.__dict__[name]

ContextGroup._STANDARD_ATTRS = set(dir(ContextGroup()))




# encoding: utf-8

"""Diesel-based WSGI server adapter."""

# ## Imports

from __future__ import unicode_literals, print_function

from diesel.protocols.wsgi import WSGIApplication


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080):
	"""Diesel-based (greenlet) WSGI-HTTP server.
	
	As a minor note, this is crazy. Diesel includes Flask, too.
	"""
	
	# Instantiate the server with a host/port configuration and our application.
	WSGIApplication(application, port=int(port), iface=host).run()




# encoding: utf-8

"""A production quality flup-based FastCGI server."""

# ## Imports

from __future__ import unicode_literals, print_function

try:
	from flup.server.fcgi import WSGIServer
except ImportError:
	print("You must install a 'flup' package such as 'flup6' to use FastCGI support.")
	raise


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080, socket=None, **options):
	"""Basic FastCGI support via flup.
	
	This web server has many, many options. Please see the Flup project documentation for details.
	"""
	
	# Allow either on-disk socket (recommended) or TCP/IP socket use.
	if not socket:
		bindAddress = (host, int(port))
	else:
		bindAddress = socket
	
	# Bind and start the blocking web server interface.
	WSGIServer(application, bindAddress=bindAddress, **options).run()




# encoding: utf-8

# ## Imports

from __future__ import unicode_literals

from collections import deque
from inspect import isclass
from marrow.package.host import PluginManager


# ## Module Globals

# A standard logger object.
log = __import__('logging').getLogger(__name__)


# ## Dispatch Plugin Manager

class WebDispatchers(PluginManager):
	"""WebCore dispatch protocol adapter.
	
	The process of resolving a request path to an endpoint. The dispatch protocol utilized is documented in full
	in the `protocol` project: https://github.com/marrow/protocols/blob/master/dispatch/README.md
	
	The allowable controller structures offered by individual methods of dispatch is documented in the relevant
	dispatch project. Examples of dispatch include:
	
	* Object Dispatch: https://github.com/marrow/web.dispatch.object
	* Registered Routes: https://github.com/marrow/web.dispatch.route
	* Traversal: https://github.com/marrow/web.dispatch.traversal
	
	Others may exist, and dispatch middleware may be available to perform more complex behaviours. The default
	dispatcher if not otherwise configured is object dispatch.
	"""
	
	__isabstractmethod__ = False  # Work around an issue in modern (3.4+) Python due to our instances being callable.
	
	def __init__(self, ctx):
		"""Dispatch registry constructor.
		
		The dispatch registry is not meant to be instantiated by third-party software. Instead, access the registry as
		an attribute of the current Application or Request context: `context.dispatch`
		"""
		
		super(WebDispatchers, self).__init__('web.dispatch')
	
	def __call__(self, context, handler, path):
		"""Having been bound to an appropriate context, find a handler for the request path.
		
		This is the part of the WebCore request cycle that speaks the Dispatch protocol and performs event bridging.
		
		This requires a context prepared with a `context.extension.signal.dispatch` list and dispatch plugin registry
		as `context.dispatch`.  This does not use `self` to allow for more creative overriding.
		"""
		
		is_endpoint = False  # We'll search until we find an endpoint.
		
		# This technically doesn't help Pypy at all, but saves repeated deep lookup in CPython.(E)
		callbacks = context.extension.signal.dispatch  # These extensions want to be notified.
		self = context.dispatch  # Dispatch plugin registry.
		
		if __debug__:
			log.debug("Preparing dispatch.", extra=dict(
					request = id(context.request),
					path = context.request.path_info,
					handler = repr(handler)
				))
		
		# Now we need the remaining path elements as a deque.
		path = path.strip('/')
		path = deque(path.split('/')) if path else deque()
		
		try:
			while not is_endpoint:
				# Pull the dispatcher out of the current handler, defaulting to object dispatch.
				dispatcher = self[getattr(handler, '__dispatch__', 'object')]
				
				# We don't stop if the same dispatcher is loaded twice since some dispatchers might want to do that.
				starting = handler
				
				# Iterate dispatch events, issuing appropriate callbacks as we descend.
				for consumed, handler, is_endpoint in dispatcher(context, handler, path):
					if is_endpoint and not callable(handler) and hasattr(handler, '__dispatch__'):
						is_endpoint = False
					# DO NOT add production logging statements (ones not wrapped in `if __debug__`) to this callback!
					for ext in callbacks: ext(context, consumed, handler, is_endpoint)
				
				# Repeat of earlier, we do this after extensions in case anything above modifies the environ path.
				path = context.environ['PATH_INFO'].strip('/')
				path = deque(path.split('/')) if path else deque()
				
				if not is_endpoint and starting is handler:  # We didn't go anywhere.
					break
		
		# Dispatch failed utterly.
		except LookupError:
			pass  # `is_endpoint` can only be `False` here.
		
		return is_endpoint, handler if is_endpoint else None
	
	def __getitem__(self, dispatcher):
		"""Retrieve a dispatcher from the registry.
		
		This performs some additional work beyond the standard plugin manager in order to construct configured
		instances of the dispatchers instead of simply returning them bare. This allows for configuration and caching
		of these configured dispatchers to happen in a single place.
		"""
		
		name = None
		
		if callable(dispatcher) and not isclass(dispatcher):
			return dispatcher
		
		# If the dispatcher isn't already executable, it's probably an entry point reference. Load it from cache.
		if not isclass(dispatcher):
			name = dispatcher
			dispatcher = super(WebDispatchers, self).__getitem__(dispatcher)
		
		# If it's uninstantiated, instantiate it.
		if isclass(dispatcher):
			# TODO: Extract **kw settings from context.
			dispatcher = dispatcher()  # Instantiate the dispatcher.
			if name:  # Update the entry point cache if loaded by name.
				self.named[name] = dispatcher
		
		if __debug__:
			log.debug("Loaded dispatcher.", extra=dict(dispatcher=repr(dispatcher)))
		
		return dispatcher




# encoding: utf-8

"""Primary WSGI application and framework entry point."""

# ## Imports

from __future__ import unicode_literals

import logging
import logging.config

from inspect import isfunction
from webob.exc import HTTPException, HTTPNotFound, HTTPInternalServerError
from marrow.package.loader import load

from .context import Context
from .dispatch import WebDispatchers
from .extension import WebExtensions
from .view import WebViews
from ..ext.base import BaseExtension
from ..ext import args as arguments

if __debug__:
	from .util import safe_name


# ## Module Globals

# A standard Python logger object.
log = __import__('logging').getLogger(__name__)


# ## WSGI Application

class Application(object):
	"""The WebCore WSGI application.
	
	This glues together a few components:
	
	* Loading and preparation the Application configuration.
	* Simple or verbose logging configuration.
	* Collection and execution of `web.extension` callbacks.
	* WSGI middleware wrapping.
	* The final WSGI application handling requests.
	"""
	
	__slots__ = (
			'config',  # Application configuration.
			'feature',  # Feature tag announcement; populated by the `provides` of active extensions.
			
			'__context',  # Application context instance.
			'RequestContext',  # Per-request context class.
			'__call__',  # WSGI request handler.  Dynamically assigned.
		)
	
	def __init__(self, root, **config):
		"""Construct the initial ApplicationContext, populate, and prepare the WSGI stack.
		
		No actions other than configuration should happen during construction.
		
		Current configuration is limited to three arguments:
		
		* `root` -- the object to use as the starting point of dispatch on each request
		* `logging` -- either `None` to indicate WebCore should not manipulate the logging configuration (the
			default), a string representing the logging level to globally configure (such as `"debug"`), or a
			dictionary configuration to pass to the Python standard `logging.dictConfig()` process.
		* `extensions` -- a list of configured extension instances, ignoring `BaseExtension` which is automatically
			added to the extension set.
		"""
		
		self.config = self._configure(config)  # Prepare the configuration.
		
		if __debug__:
			log.debug("Preparing WebCore application.")
		
		if isfunction(root):  # We need to armour against this turning into a bound method of the context.
			root = staticmethod(root)
		
		# This construts a basic `ApplicationContext` containing a few of the passed-in values.
		context = self.__context = Context(app=self, root=root)._promote('ApplicationContext')
		
		# These can't really be deferred to extensions themselves, for fairly obvious chicken/egg reasons.
		exts = context.extension = WebExtensions(context)  # Load extension registry and prepare callbacks.
		context.dispatch = WebDispatchers(context)  # Load dispatch registry.
		context.view = WebViews(context)  # Load the view registry.
		
		# Execute extension startup callbacks; this is the appropriate time to attach descriptors to the context.
		for ext in exts.signal.start: ext(context)
		
		# At this point the context should have been populated with any descriptor protocol additions. Promote the
		# `ApplicationContext` instance to a `RequestContext` class for use during the request/response cycle.
		self.RequestContext = context._promote('RequestContext', instantiate=False)
		
		# Handle WSGI middleware wrapping by extensions and point our __call__ at the result.
		app = self.application
		for ext in exts.signal.middleware: app = ext(context, app)
		self.__call__ = app
		
		if __debug__:  # Mostly useful for timing calculations.
			log.debug("WebCore application prepared.")
	
	def _configure(self, config):
		"""Prepare the incoming configuration and ensure certain expected values are present.
		
		For example, this ensures BaseExtension is included in the extension list, and populates the logging config.
		"""
		config = config or dict()
		
		# We really need this to be there.
		if 'extensions' not in config: config['extensions'] = list()
		
		if not any(isinstance(ext, BaseExtension) for ext in config['extensions']):
			# Always make sure the BaseExtension is present since request/response objects are handy.
			config['extensions'].insert(0, BaseExtension())
		
		if not any(isinstance(ext, arguments.ArgumentExtension) for ext in config['extensions']):
			# Prepare a default set of argument mutators.
			config['extensions'].extend([
					arguments.ValidateArgumentsExtension(),
					arguments.ContextArgsExtension(),
					arguments.RemainderArgsExtension(),
					arguments.QueryStringArgsExtension(),
					arguments.FormEncodedKwargsExtension(),
					arguments.JSONKwargsExtension(),
				])
		
		# Tests are skipped on these as we have no particular need to test Python's own logging mechanism.
		level = config.get('logging', {}).get('level', None)
		if level:  # pragma: no cover
			logging.basicConfig(level=getattr(logging, level.upper()))
		elif 'logging' in config:  # pragma: no cover
			logging.config.dictConfig(config['logging'])
		
		return config
	
	# This is impractical to test due to the blocking nature of starting a web server interface.
	# Individual adapters are hand-tested for basic operation prior to release.
	def serve(self, service='auto', **options):  # pragma: no cover
		"""Initiate a web server service to serve this application.
		
		You can always use the Application instance as a bare WSGI application, of course.  This method is provided as
		a convienence.
		
		Pass in the name of the service you wish to use, and any additional configuration options appropriate for that
		service.  Almost all services accept `host` and `port` options, some also allow you to specify an on-disk
		`socket`.  By default all web servers will listen to `127.0.0.1` (loopback only) on port 8080.
		"""
		
		service = load(service, 'web.server')  # We don't bother with a full registry for these one-time lookups.
		
		try:
			service(self, **options)
		except KeyboardInterrupt:  # We catch this as SIG_TERM or ^C are basically the only ways to stop most servers.
			pass
		
		# Notify extensions that the service has returned and we are exiting.
		for ext in self.__context.extension.signal.stop: ext(self.__context)
	
	def _execute_endpoint(self, context, endpoint, signals):
		if not callable(endpoint):
			# Endpoints don't have to be functions.
			# They can instead point to what a function would return for view lookup.
			
			if __debug__:
				log.debug("Static endpoint located.", extra=dict(
						request = id(context),
						endpoint = repr(endpoint),
					))
			
			# Use the result directly, as if it were the result of calling a function or method.
			return endpoint
		
		# Populate any endpoint arguments and allow for chained mutation and validation.
		args, kwargs = [], {}
		
		try:
			for ext in signals.mutate: ext(context, endpoint, args, kwargs)
		
		except HTTPException as e:
			result = e
		
		else:
			# If successful in accumulating arguments, finally call the endpoint.
			
			if __debug__:
				log.debug("Callable endpoint located and arguments prepared.", extra=dict(
						request = id(context),
						endpoint = safe_name(endpoint),
						endpoint_args = args,
						endpoint_kw = kwargs
					))
			
			try:
				result = endpoint(*args, **kwargs)
			
			except HTTPException as e:
				result = e
		
		# Execute return value transformation callbacks.
		for ext in signals.transform: result = ext(context, endpoint, result)
		
		return result
	
	def application(self, environ, start_response):
		"""Process a single WSGI request/response cycle.
		
		This is the WSGI handler for WebCore.  Depending on the presence of extensions providing WSGI middleware,
		the `__call__` attribute of the Application instance will either become this, or become the outermost
		middleware callable.
		
		Most apps won't utilize middleware, the extension interface is preferred for most operations in WebCore.
		They allow for code injection at various intermediary steps in the processing of a request and response.
		"""
		context = environ['wc.context'] = self.RequestContext(environ=environ)
		signals = context.extension.signal
		
		# Announce the start of a request cycle. This executes `prepare` and `before` callbacks in the correct order.
		for ext in signals.pre: ext(context)
		
		# Identify the endpoint for this request.
		is_endpoint, handler = context.dispatch(context, context.root, context.environ['PATH_INFO'])
		
		if is_endpoint:
			try:
				result = self._execute_endpoint(context, handler, signals)  # Process the endpoint.
			except Exception as e:
				log.exception("Caught exception attempting to execute the endpoint.")
				result = HTTPInternalServerError(str(e) if __debug__ else "Please see the logs.")
				
				if 'debugger' in context.extension.feature:
					context.response = result
					for ext in signals.after: ext(context)  # Allow signals to clean up early.
					raise
		
		else:  # If no endpoint could be resolved, that's a 404.
			result = HTTPNotFound("Dispatch failed." if __debug__ else None)
		
		if __debug__:
			log.debug("Result prepared, identifying view handler.", extra=dict(
					request = id(context),
					result = safe_name(type(result))
				))
		
		# Identify a view capable of handling this result.
		for view in context.view(result):
			if view(context, result): break
		else:
			# We've run off the bottom of the list of possible views.
			raise TypeError("No view could be found to handle: " + repr(type(result)))
		
		if __debug__:
			log.debug("View identified, populating response.", extra=dict(
					request = id(context),
					view = repr(view),
				))
		
		for ext in signals.after: ext(context)
		
		def capture_done(response):
			for chunk in response:
				yield chunk
			
			for ext in signals.done: ext(context)
		
		# This is really long due to the fact we don't want to capture the response too early.
		# We need anything up to this point to be able to simply replace `context.response` if needed.
		return capture_done(context.response.conditional_response_app(environ, start_response))




# encoding: utf-8

# ## Imports

from __future__ import unicode_literals

from threading import local

from marrow.package.loader import traverse


# ## Module Globals

# Standard logger object.
log = __import__('logging').getLogger(__name__)


# ## Extension

class ThreadLocalExtension(object):
	"""Provide the current context as a thread local global.
	
	This provides a convienent "superglobal" variable where you can store per-thread data.
	
	While the context itself is cleaned up after each call, any data you add won't be.  These are not request-locals.
	"""
	
	first = True
	provides = ['local', 'threadlocal']
	
	def __init__(self, where='web.core:local'):
		"""Initialize thread local storage for the context.
		
		By default the `local` object in the `web.core` package will be populated as a `threading.local` pool. The
		context, during a request, can then be accessed as `web.core.local.context`. Your own extensions can add
		additional arbitrary data to this pool.
		"""
		
		super(ThreadLocalExtension, self).__init__()
		
		if __debug__:
			log.debug("Initalizing ThreadLocal extension.")
		
		self.where = where
		self.local = None
		self.preserve = False
	
	def _lookup(self):
		module, _, name = self.where.rpartition(':')
		module = traverse(__import__(module), '.'.join(module.split('.')[1:]), separator='.')
		
		return module, name
	
	def start(self, context):
		module, name = self._lookup()
		
		if __debug__:
			log.debug("Preparing thread local storage and assigning main thread application context.")
		
		if hasattr(module, name):
			self.local = getattr(module, name)
			self.preserve = True
		else:
			self.local = local()
			setattr(module, name, self.local)
		
		self.local.context = context  # Main thread application context.
	
	def stop(self, context):
		self.local = None
		
		if __debug__:
			log.debug("Cleaning up thread local storage.")
		
		if not self.preserve:
			module, name = self._lookup()
			delattr(module, name)
	
	def prepare(self, context):
		"""Executed prior to processing a request."""
		if __debug__:
			log.debug("Assigning thread local request context.")
		
		self.local.context = context
	
	def done(self, result):
		"""Executed after the entire response has been sent to the client."""
		if __debug__:
			log.debug("Cleaning up thread local request context.")
		
		del self.local.context




# encoding: utf-8

"""Python 3 function annotation typecasting support."""

# ## Imports

from __future__ import unicode_literals

from inspect import ismethod, getfullargspec

from web.core.compat import items


# ## Extension

class AnnotationExtension(object):
	"""Utilize Python 3 function annotations as a method to filter arguments coming in from the web.
	
	Argument annotations are treated as callbacks to execute, passing in the unicode value coming in from the web and
	swapping it with the value returned by the callback. This allows for trivial typecasting to most built-in Python
	types such as `int`, `float`, etc., as well as creative use such as `','.split` to automatically split a comma-
	separated value. One can of course also write custom callbacks, notably ones that raise `HTTPException`
	subclasses to not appear as an Internal Server Error.
	
	For example:
	
		def multiply(a: int, b: int):
			return str(a * b)
	
	This extension also performs a utility wrapping of returned values in the form of a 2-tuple of the return
	annotation itself and the value returned by the callable endpoint. This integrates well with the view registered
	by the `web.template` package to define a template at the head of the function, returning data for the template
	to consume:
	
		def hello(name="world"): -> 'mako:hello.html'
			return dict(name=name)
	
	If your editor has difficulty syntax highlighting such annotations, check for a Python 3 compatible update to your
	editor's syntax definitions.
	"""
	
	__slots__ = tuple()
	
	provides = ['annotation', 'cast', 'typecast']  # Export these symbols for other extensions to depend upon.
	
	# ### Request-Local Callbacks
	
	def mutate(self, context, handler, args, kw):
		"""Inspect and potentially mutate the given handler's arguments.
		
		The args list and kw dictionary may be freely modified, though invalid arguments to the handler will fail.
		"""
		annotations = getattr(handler.__func__ if hasattr(handler, '__func__') else handler, '__annotations__', None)
		if not annotations:
			return
		
		argspec = getfullargspec(handler)
		arglist = list(argspec.args)
		
		if ismethod(handler):
			del arglist[0]
		
		for i, value in enumerate(list(args)):
			key = arglist[i]
			if key in annotations:
				args[i] = annotations[key](value)
		
		# Convert keyword arguments
		for key, value in list(items(kw)):
			if key in annotations:
				kw[key] = annotations[key](value)
	
	def transform(self, context, handler, result):
		"""Transform the value returned by the controller endpoint.
		
		This extension transforms returned values if the endpoint has a return type annotation.
		"""
		handler = handler.__func__ if hasattr(handler, '__func__') else handler
		annotation = getattr(handler, '__annotations__', {}).get('return', None)
		
		if annotation:
			return (annotation, result)
		
		return result




# encoding: utf-8

"""Python-standard reference servers for development use."""

# ## Imports

from __future__ import unicode_literals, print_function

from wsgiref.handlers import CGIHandler
from wsgiref.simple_server import make_server


# ## Production Warning

# We let people know it's a bad idea to use these in production.
if not __debug__:
	import warnings
	warnings.warn("Use of standard library reference servers in production is discouraged.", RuntimeWarning)


# ## Server Adapters

def simple(application, host='127.0.0.1', port=8080):
	"""Python-standard WSGI-HTTP server for testing purposes.
	
	The additional work performed here is to match the default startup output of "waitress".
	
	This is not a production quality interface and will be have badly under load.
	"""
	
	# Try to be handy as many terminals allow clicking links.
	print("serving on http://{0}:{1}".format(host, port))
	
	# Bind and launch the server; this is a blocking operation.
	make_server(host, int(port), application).serve_forever()


def cgi(application):
	"""Python-standard WSGI-CGI server for testing purposes.
	
	This is not a production quality interface and will behave badly under load. Python-as-CGI is not a very good way
	to deploy any application. (Startup/shutdown on every request is a PHP problem.) This _can_ be useful as a
	diagnostic tool in development, however.
	"""
	
	if not __debug__:
		warnings.warn("Interactive debugging and other persistence-based processes will not work.")
	
	# Instantiate the handler and begin bridging the application.
	CGIHandler().run(application)


def iiscgi(application):
	"""A specialized version of the reference WSGI-CGI server to adapt to Microsoft IIS quirks.
	
	This is not a production quality interface and will behave badly under load.
	"""
	try:
		from wsgiref.handlers import IISCGIHandler
	except ImportError:
		print("Python 3.2 or newer is required.")
	
	if not __debug__:
		warnings.warn("Interactive debugging and other persistence-based processes will not work.")
	
	IISCGIHandler().run(application)






# encoding: utf-8

"""Compatibility helpers to bridge the differences between Python 2 and Python 3.

Similar in purpose to [`six`](https://warehouse.python.org/project/six/). Not generally intended to be used by
third-party software, these are subject to change at any time. Only symbols exported via `__all__` are safe to use.
"""

# ## Imports

import sys


# ## Module Exports

__all__ = ['py3', 'pypy', 'unicode', 'str']


# ## Version Detection

py3 = sys.version_info > (3, )
pypy = hasattr(sys, 'pypy_version_info')


# ## Builtins Compatibility

# Use of the `items` shortcut here must be very, very careful to only apply it to actual bare dictionaries.

if py3:
	unicode = str
	str = bytes
	items = dict.items
else:
	unicode = unicode
	str = str
	items = dict.iteritems


# ## File-Like String Handling

try:
	try:
		from cStringIO import StringIO
	except ImportError:
		from StringIO import StringIO
except ImportError:
	from io import StringIO


# ## Python Standard Library Backports

try:
	from pathlib import PurePosixPath as Path
except ImportError:
	from pathlib2 import PurePosixPath as Path




# encoding: utf-8

"""An example, though quite usable extension to handle list and dictionary return values."""

# ## Imports

from __future__ import unicode_literals

import pkg_resources
from collections import Mapping

from marrow.package.host import PluginManager
from web.core.compat import str


try:
	from bson import json_util as json
except ImportError:
	import json


# ## Module Globals

log = __import__('logging').getLogger(__name__)
json  # Satisfy linter.


# ## Plugin Management


class SerializationPlugins(PluginManager):
	def __init__(self, namespace, folders=None):
		self.__dict__['names'] = set()
		self.__dict__['types'] = set()
		super(SerializationPlugins, self).__init__(namespace, folders)
	
	def register(self, name, plugin):
		super(SerializationPlugins, self).register(name, plugin)
		
		self.names.add(name)
		
		if '/' in name:
			self.types.add(name)
	
	def _register(self, dist):
		try:
			super(SerializationPlugins, self)._register(dist)
		except pkg_resources.DistributionNotFound:
			pass


# ## Extension

class SerializationExtension(object):
	"""Sample extension demonstrating integration of automatic serialization, such as JSON.
	
	This extension registers handlers for lists and dictionaries (technically list and mappings).
	
	Additional serializers can be registered during runtime by other extensions by adding a new mimetype mapping
	to the `context.serialize` dictionary-like object.  For convienence the default serializers are also provided
	using their simple names, so you can access the JSON encoder directly, for example:
	
		context.serialize.json.dumps(...)
	"""
	
	provides = {'serialization'}
	extensions = {'web.serializer'}
	context = {'serialize'}
	
	def __init__(self, default='application/json', types=(list, Mapping)):
		self.default = default
		self.types = types
	
	# ### Application-Level Callbacks
	
	def start(self, context):
		if __debug__:
			log.debug("Registering serialization return value handlers.")
		
		manager = SerializationPlugins('web.serialize')
		manager.__dict__['__isabstractmethod__'] = False
		
		context.serialize = manager
		
		# Register the serialization views supported by this extension.
		for kind in self.types:
			context.view.register(kind, self.render_serialization)
	
	# ### Views
	
	def render_serialization(self, context, result):
		"""Render serialized responses."""
		
		resp = context.response
		serial = context.serialize
		match = context.request.accept.best_match(serial.types, default_match=self.default)
		result = serial[match](result)
		
		if isinstance(result, str):
			result = result.decode('utf-8')
		
		resp.charset = 'utf-8'
		resp.content_type = match
		resp.text = result
		
		return True




# encoding: utf-8

"""Eventlet-based WSGI server adapter."""

# ## Imports

from __future__ import unicode_literals, print_function

from eventlet import listen
from eventlet.wsgi import server


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080):
	"""Eventlet-based WSGI-HTTP server.
	
	For a more fully-featured Eventlet-capable interface, see also [Spawning](http://pypi.python.org/pypi/Spawning/).
	"""
	
	# Instantiate the server with a bound port and with our application.
	server(listen(host, int(port)), application)




# encoding: utf-8

# ## Imports

from __future__ import unicode_literals, print_function

try:
	import tornado.ioloop
	import tornado.httpserver
	import tornado.wsgi
except ImportError:
	print("You must install the 'tornado' package.")
	raise


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080, **options):
	"""Tornado's HTTPServer.
	
	This is a high quality asynchronous server with many options.  For details, please visit:
	
		http://www.tornadoweb.org/en/stable/httpserver.html#http-server
	"""
	
	# Wrap our our WSGI application (potentially stack) in a Tornado adapter.
	container = tornado.wsgi.WSGIContainer(application)
	
	# Spin up a Tornado HTTP server using this container.
	http_server = tornado.httpserver.HTTPServer(container, **options)
	http_server.listen(int(port), host)
	
	# Start and block on the Tornado IO loop.
	tornado.ioloop.IOLoop.instance().start()




# encoding: utf-8

"""Argument handling extensions for WebCore applications.

These allow you to customize the behaviour of the arguments passed to endpoints.
"""

from inspect import isroutine, ismethod, getcallargs

from webob.exc import HTTPNotFound
from web.core.util import safe_name


# A standard Python logger object.
log = __import__('logging').getLogger(__name__)


class ArgumentExtension(object):
	"""Not for direct use."""
	
	@staticmethod
	def _process_flat_kwargs(source, kwargs):
		"""Apply a flat namespace transformation to recreate (in some respects) a rich structure.
		
		This applies several transformations, which may be nested:
		
		`foo` (singular): define a simple value named `foo`
		`foo` (repeated): define a simple value for placement in an array named `foo`
		`foo[]`: define a simple value for placement in an array, even if there is only one
		`foo.<id>`: define a simple value to place in the `foo` array at the identified index
		
		By nesting, you may define deeper, more complex structures:
		
		`foo.bar`: define a value for the named element `bar` of the `foo` dictionary
		`foo.<id>.bar`: define a `bar` dictionary element on the array element marked by that ID
		
		References to `<id>` represent numeric "attributes", which makes the parent reference be treated as an array,
		not a dictionary. Exact indexes might not be able to be preserved if there are voids; Python lists are not
		sparse.
		
		No validation of values is performed.
		"""
		
		ordered_arrays = []
		
		# Process arguments one at a time and apply them to the kwargs passed in.
		
		for name, value in source.items():
			container = kwargs
			
			if '.' in name:
				parts = name.split('.')
				name = name.rpartition('.')[2]
				
				for target, following in zip(parts[:-1], parts[1:]):
					if following.isnumeric():  # Prepare any use of numeric IDs.
						container.setdefault(target, [{}])
						if container[target] not in ordered_arrays:
							ordered_arrays.append(container[target])
						container = container[target][0]
						continue
					
					container = container.setdefault(target, {})
			
			if name.endswith('[]'):  # `foo[]` or `foo.bar[]` etc.
				name = name[:-2]
				container.setdefault(name, [])
				container[name].append(value)
				continue
			
			if name.isnumeric() and container is not kwargs:  # trailing identifiers, `foo.<id>`
				container[int(name)] = value
				continue
			
			if name in container:
				if not isinstance(container[name], list):
					container[name] = [container[name]]
				
				container[name].append(value)
				continue
			
			container[name] = value
		
		for container in ordered_arrays:
			elements = container[0]
			del container[:]
			container.extend(value for name, value in sorted(elements.items()))
	
	@staticmethod
	def _process_rich_kwargs(source, kwargs):
		"""Apply a nested structure to the current kwargs."""
		kwargs.update(source)


class ValidateArgumentsExtension(object):
	"""Use this to enable validation of endpoint arguments.
	
	You can determine when validation is executed (never, always, or development) and what action is taken when a
	conflict occurs.
	"""
	
	last = True
	
	provides = {'args.validation', 'kwargs.validation'}
	
	def __init__(self, enabled='development', correct=False):
		"""Configure when validation is performed and the action performed.
		
		If `enabled` is `True` validation will always be performed, if `False`, never. If set to `development` the
		callback will not be assigned and no code will be executed during runtime.
		
		When `correct` is falsy (the default), an `HTTPNotFound` will be raised if a conflict occurs. If truthy the
		conflicting arguments are removed, with positional taking precedence to keyword.
		"""
		
		if enabled is True or (enabled == 'development' and __debug__):
			self.mutate = self._mutate
	
	def _mutate(self, context, endpoint, args, kw):
		try:
			if callable(endpoint) and not isroutine(endpoint):
				endpoint = endpoint.__call__  # Handle instances that are callable.
			
			getcallargs(endpoint, *args, **kw)
		
		except TypeError as e:
			# If the argument specification doesn't match, the handler can't process this request.
			# This is one policy. Another possibility is more computationally expensive and would pass only
			# valid arguments, silently dropping invalid ones. This can be implemented as a mutate handler.
			log.error(str(e).replace(endpoint.__name__, safe_name(endpoint)), extra=dict(
					request = id(context),
					endpoint = safe_name(endpoint),
					endpoint_args = args,
					endpoint_kw = kw,
				))
			
			raise HTTPNotFound("Incorrect endpoint arguments: " + str(e))


class ContextArgsExtension(ArgumentExtension):
	"""Add the context as the first positional argument, possibly conditionally."""
	
	first = True
	provides = {'args.context'}
	
	def __init__(self, always=False):
		"""Configure the conditions under which the context is added to endpoint positional arguments.
		
		When `always` is truthy the context is always included, otherwise it's only included for callables that are
		not bound methods.
		"""
		self.always = always
	
	def mutate(self, context, endpoint, args, kw):
		if not self.always:
			# Instance methods were handed the context at class construction time via dispatch.
			# The `not isroutine` bit here catches callable instances, a la "index.html" handling.
			if not isroutine(endpoint) or (ismethod(endpoint) and getattr(endpoint, '__self__', None) is not None):
				return
		
		args.insert(0, context)


class RemainderArgsExtension(ArgumentExtension):
	"""Add any unprocessed path segments as positional arguments."""
	
	first = True
	needs = {'request'}
	uses = {'args.context'}
	provides = {'args.remainder'}
	
	def mutate(self, context, endpoint, args, kw):
		if not context.request.remainder:
			return
		
		args.extend(context.request.remainder)


class QueryStringArgsExtension(ArgumentExtension):
	"""Add query string arguments ("GET") as keyword arguments."""
	
	first = True
	needs = {'request'}
	provides = {'kwargs.get'}
	
	def mutate(self, context, endpoint, args, kw):
		self._process_flat_kwargs(context.request.GET, kw)


class FormEncodedKwargsExtension(ArgumentExtension):
	"""Add form-encoded or MIME mmultipart ("POST") arguments as keyword arguments."""
	
	first = True
	needs = {'request'}
	uses = {'kwargs.get'}  # Query string values must be processed first, to be overridden.
	provides = {'kwargs.post'}
	
	def mutate(self, context, endpoint, args, kw):
		self._process_flat_kwargs(context.request.POST, kw)


class JSONKwargsExtension(ArgumentExtension):
	"""Add JSON-encoded arguments from the request body as keyword arguments."""
	
	first = True
	needs = {'request'}
	uses = {'kwargs.get'}  # We override values defined in the query string.
	provides = {'kwargs.json'}
	
	def mutate(self, context, endpoint, args, kw):
		if not context.request.content_type == 'application/json':
			return
		
		if not context.request.body:
			return
		
		self._process_rich_kwargs(context.request.json, kw)





# encoding: utf-8

"""CherryPy-based WSGI server adapter."""

# ## Imports

from __future__ import unicode_literals, print_function

from cherrypy.wsgiserver import CherryPyWSGIServer


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080):
	"""CherryPy-based WSGI-HTTP server."""
	
	# Instantiate the server with our configuration and application.
	server = CherryPyWSGIServer((host, int(port)), application, server_name=host)
	
	# Try to be handy as many terminals allow clicking links.
	print("serving on http://{0}:{1}".format(host, port))
	
	# Bind and launch the server; this is a blocking operation.
	try:
		server.start()
	except KeyboardInterrupt:
		server.stop()  # CherryPy has some of its own shutdown work to do.




# encoding: utf-8

"""Record basic performance statistics."""

# ## Imports

from __future__ import unicode_literals

import time

from web.core.compat import unicode


# ## Module Globals

log = __import__('logging').getLogger(__name__)


# ## Extension

class AnalyticsExtension(object):
	"""Record performance statistics about each request, and potentially a lot more.
	
	By default this extension adds a `X-Generation-Time` header to all responses and logs the generation time at the
	`debug` level.  You can disable either by passing `header=None` or `level=None`, or specify an alternate logging
	level by passing in the name of the level.
	"""
	
	__slots__ = ('header', 'log')
	
	first = True  # We need this processing to happen as early as possible.
	provides = ['analytics']  # Expose this symbol for other extensions to depend upon.
	
	def __init__(self, header='X-Generation-Time', level='debug'):
		"""Executed to configure the extension."""
		
		super(AnalyticsExtension, self).__init__()
		
		# Record settings.
		self.header = header
		self.log = getattr(log, level) if level else None
	
	# ### Request-Local Callabacks
	
	def prepare(self, context):
		"""Executed during request set-up."""
		
		context._start_time = None
	
	def before(self, context):
		"""Executed after all extension prepare methods have been called, prior to dispatch."""
		
		context._start_time = time.time()
	
	def after(self, context, exc=None):
		"""Executed after dispatch has returned and the response populated, prior to anything being sent to the client."""
		
		duration = time.time() - context._start_time
		delta = unicode(round(duration, 5))
		
		# Default response augmentation.
		if self.header:
			context.response.headers[self.header] = delta
		
		if self.log:
			self.log("Response generated in " + delta + " seconds.", extra=dict(
					duration = duration,
					request = id(context)
				))




# encoding: utf-8

"""The recommended development HTTP server."""

# ## Imports

from __future__ import unicode_literals, print_function

try:
	from waitress import serve as serve_
except ImportError:
	print("You must install the 'waitress' package.")
	raise


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080, threads=4, **kw):
	"""The recommended development HTTP server.
	
	Note that this server performs additional buffering and will not honour chunked encoding breaks.
	"""
	
	# Bind and start the server; this is a blocking process.
	serve_(application, host=host, port=int(port), threads=int(threads), **kw)




# encoding: utf-8

"""Python-standard reference servers for development use."""

# ## Imports

from __future__ import unicode_literals

import warnings

from google.appengine.ext.webapp.util import run_wsgi_app


# ## Server Adapter

def appengine(application):
	"""Google App Engine adapter, CGI.
	
	Note: This adapter is essentially untested, and likely duplicates the `cgiref` adapter.
	"""
	
	warnings.warn("Interactive debugging and other persistence-based processes will not work.")
	
	# Bridge the current CGI request.
	run_wsgi_app(application)




# encoding: utf-8

"""Release information about WebCore."""

# ## Imports

from __future__ import unicode_literals

import sys
from collections import namedtuple


# ## Module Globals

version_info = namedtuple('version_info', ('major', 'minor', 'micro', 'releaselevel', 'serial'))(2, 0, 4, 'final', 0)
version = ".".join([str(i) for i in version_info[:3]]) + ((version_info.releaselevel[0] + str(version_info.serial)) if version_info.releaselevel != 'final' else '')

author = namedtuple('Author', ['name', 'email'])("Alice Bevan-McGregor", 'alice@gothcandy.com')
description = "A powerful web development nanoframework so small it's not even a microframework."
copyright = "2009-2020, Alice Bevan-McGregor and contributors"
url = 'https://github.com/marrow/WebCore/'
colophon = """Powered by:
	<a class="noicon" href="http://www.python.org/" title="A programming language that lets you work more quickly and integrate your systems more effectively.">Python</a> <em class="version">{0.major}.{0.minor}</em>
	and <a class="noicon" href="{1}" title="A lightweight and extremely fast Python web framework.">WebCore</a> <em class="version">{2}</em>.""".format(sys.version_info, url, version)




# encoding: utf-8

"""Gevent-based WSGI server adapter."""

# ## Imports

from __future__ import unicode_literals, print_function

from gevent.pywsgi import WSGIServer


# ## Server Adapter

def serve(application, host='127.0.0.1', port=8080):
	"""Gevent-based WSGI-HTTP server."""
	
	# Instantiate the server with a host/port configuration and our application.
	WSGIServer((host, int(port)), application).serve_forever()




# encoding: utf-8

# ## Imports

from threading import local as __local

# Expose these as importable from the top-level `web.core` namespace.

from .application import Application
from .util import lazy


# ## Module Globals

__all__ = ['local', 'Applicaiton', 'lazy']  # Symbols exported by this package.

# This is to support the web.ext.local extension, and allow for early importing of the variable.
local = __local()




# encoding: utf-8

"""The base extension providing request, response, and core views."""

# ## Imports

from __future__ import unicode_literals

from io import IOBase
try:
	IOBase = (IOBase, file)
except:
	pass

try:
	from collections import Generator
except ImportError:
	def _tmp(): yield None  # pragma: no cover
	Generator = type(_tmp())

from os.path import getmtime
from time import mktime, gmtime
from datetime import datetime
from mimetypes import init, add_type, guess_type
from collections import namedtuple
from webob import Request, Response

from web.core.compat import str, unicode, Path
from web.core.util import safe_name


# ## Module Globals

log = __import__('logging').getLogger(__name__)


# ## Helper Classes

Crumb = namedtuple('Breadcrumb', ('handler', 'path'))


class Bread(list):
	@property
	def current(self):
		return self[-1].path



# ## Extension

class BaseExtension(object):
	"""Base framework extension.
	
	This extension is not meant to be manually constructed or manipulated; use is automatic.
	"""
	
	first = True  # Must occur as early as possible in callback lists.
	always = True  # Always enabled.
	provides = ["base", "request", "response"]  # Export these symbols for use as other extension's dependencies.
	
	# ### Application-Level Callbacks
	
	def start(self, context):
		if __debug__:
			log.debug("Registering core return value handlers.")
		
		# This prepares the mimetypes registry, and adds values typically missing from it.
		init()
		add_type('text/x-yaml', 'yml')
		add_type('text/x-yaml', 'yaml')
		
		# Register the core views supported by the base framework.
		register = context.view.register
		register(type(None), self.render_none)
		register(Response, self.render_response)
		register(str, self.render_binary)
		register(unicode, self.render_text)
		register(IOBase, self.render_file)
		register(Generator, self.render_generator)
	
	# ### Request-Level Callbacks
	
	def prepare(self, context):
		"""Add the usual suspects to the context.
		
		This adds `request`, `response`, and `path` to the `RequestContext` instance.
		"""
		
		if __debug__:
			log.debug("Preparing request context.", extra=dict(request=id(context)))
		
		# Bridge in WebOb `Request` and `Response` objects.
		# Extensions shouldn't rely on these, using `environ` where possible instead.
		context.request = Request(context.environ)
		context.response = Response(request=context.request)
		
		# Record the initial path representing the point where a front-end web server bridged to us.
		context.environ['web.base'] = context.request.script_name
		
		# Track the remaining (unprocessed) path elements.
		context.request.remainder = context.request.path_info.strip('/').split('/')
		if context.request.remainder and not context.request.remainder[0]:
			del context.request.remainder[0]
		
		# Track the "breadcrumb list" of dispatch through distinct controllers.
		context.path = Bread()
	
	def dispatch(self, context, consumed, handler, is_endpoint):
		"""Called as dispatch descends into a tier.
		
		The base extension uses this to maintain the "current url".
		"""
		
		request = context.request
		
		if __debug__:
			log.debug("Handling dispatch event.", extra=dict(
					request = id(context),
					consumed = consumed,
					handler = safe_name(handler),
					endpoint = is_endpoint
				))
		
		# The leading path element (leading slash) requires special treatment.
		if not consumed and context.request.path_info_peek() == '':
			consumed = ['']
		
		nConsumed = 0
		if consumed:
			# Migrate path elements consumed from the `PATH_INFO` to `SCRIPT_NAME` WSGI environment variables.
			if not isinstance(consumed, (list, tuple)):
				consumed = consumed.split('/')
			
			for element in consumed:
				if element == context.request.path_info_peek():
					context.request.path_info_pop()
					nConsumed += 1
				else:
					break
		
		# Update the breadcrumb list.
		context.path.append(Crumb(handler, Path(request.script_name)))
		
		if consumed:  # Lastly, update the remaining path element list.
			request.remainder = request.remainder[nConsumed:]
	
	# ### Views
	
	def render_none(self, context, result):
		"""Render empty responses."""
		context.response.body = b''
		del context.response.content_length
		return True
	
	def render_response(self, context, result):
		"""Allow direct returning of WebOb `Response` instances."""
		context.response = result
		return True
	
	def render_binary(self, context, result):
		"""Return binary responses unmodified."""
		context.response.app_iter = iter((result, ))  # This wraps the binary string in a WSGI body iterable.
		return True
	
	def render_text(self, context, result):
		"""Return textual responses, encoding as needed."""
		context.response.text = result
		return True
	
	def render_file(self, context, result):
		"""Perform appropriate metadata wrangling for returned open file handles."""
		if __debug__:
			log.debug("Processing file-like object.", extra=dict(request=id(context), result=repr(result)))
		
		response = context.response
		response.conditional_response = True
		
		modified = mktime(gmtime(getmtime(result.name)))
		
		response.last_modified = datetime.fromtimestamp(modified)
		ct, ce = guess_type(result.name)
		if not ct: ct = 'application/octet-stream'
		response.content_type, response.content_encoding = ct, ce
		response.etag = unicode(modified)
		
		result.seek(0, 2)  # Seek to the end of the file.
		response.content_length = result.tell()
		
		result.seek(0)  # Seek back to the start of the file.
		response.body_file = result
		
		return True
	
	def render_generator(self, context, result):
		"""Attempt to serve generator responses through stream encoding.
		
		This allows for direct use of cinje template functions, which are generators, as returned views.
		"""
		context.response.encoding = 'utf8'
		context.response.app_iter = (
				(i.encode('utf8') if isinstance(i, unicode) else i)  # Stream encode unicode chunks.
				for i in result if i is not None  # Skip None values.
			)
		return True




# encoding: utf-8

"""WebCore extension management.

This extension registry handles loading and access to extensions as well as the collection of standard WebCore
Extension API callbacks. Reference the `SIGNALS` constant for a list of the individual callbacks that can be
utilized and their meanings, and the `extension.py` example for more detailed descriptions.

At a basic level an extension is a class. That's it; attributes and methods are used to inform the manager
of extension metadata and register callbacks for certain events. The most basic extension is one that does
nothing:

	class Extension: pass

To register your extension, add a reference to it to your project's `entry_points` in your project's `setup.py`
under the `web.extension` namespace:

	setup(
		...,
		entry_points = {'web.extension': [
				'example = myapp:Extension',
			]},
	)

Your extension may define several additional properties:

* `provides` -- declare a set of tags describing the features offered by the plugin
* `needs` -- delcare a set of tags that must be present for this extension to function
* `uses` -- declare a set of tags that must be evaluated prior to this extension, but aren't hard requirements
* `first` -- declare that this extension is a dependency of all other non-first extensions if truthy
* `last` -- declare that this extension depends on all other non-last extensions if truthy
* `signals` -- a set of additional signal names declared used (thus cacheable) by the extension manager

Tags used as `provides` values should also be registered as `web.extension` entry points. Additional `signals` may be
prefixed with a minus symbol (-) to request reverse ordering, simulating the exit path of WSGI middleware.

"""

# ## Imports

from __future__ import unicode_literals

from marrow.package.host import ExtensionManager

from .compat import items
from .context import Context


# ## Module Globals

# A standard Python logger object.
log = __import__('logging').getLogger(__name__)


# ## Extension Manager

class WebExtensions(ExtensionManager):
	"""Principal WebCore extension manager."""
	
	# Each of these is an optional extension callback attribute.
	SIGNALS = {  # Core extension hooks.
			'start',  # Executed during Application construction.
			'stop',  # Executed when (and if) the serve() server returns.
			'graceful',  # Executed when (and if) the process is instructed to reload configuration.
			'prepare',  # Executed during initial request processing.
			'dispatch',  # Executed once for each dispatch event.
			'before',  # Executed after all extension `prepare` methods have been called, prior to dispatch.
			'mutate',  # Inspect and potentially mutate arguments to the handler prior to execution.
			'-after',  # Executed after dispatch has returned and the response populated.
			'-transform',  # Transform the result returned by the handler and apply it to the response.
			'-done',  # Executed after the response has been consumed by the client.
			'-middleware',  # Executed to allow WSGI middleware wrapping.
		}
	
	__isabstractmethod__ = False  # Work around a Python 3.4+ issue when attaching to the context.
	
	# ### \_\_init__(ctx: _ApplicationContext_)
	def __init__(self, ctx):
		"""Extension registry constructor.
		
		The extension registry is not meant to be instantiated by third-party software. Instead, access the registry
		as an attribute of the current Application or Request context: `context.extension`
		
		Currently, this uses some application-internal shenanigans to construct the initial extension set.
		"""
		
		self.feature = set()  # Track the active `provides` tags.
		all = self.all = self.order(ctx.app.config['extensions'])  # Dependency ordered, active extensions.
		
		signals = {}
		inverse = set()
		
		# Prepare the known callback sets.
		
		def add_signal(name):
			if name[0] == '-':
				name = name[1:]
				inverse.add(name)
			
			signals[name] = []
		
		# Populate the initial set of signals from our own.
		for signal in self.SIGNALS: add_signal(signal)
		
		# Populate additional signals and general metadata provided by registered extensions.
		for ext in all:
			self.feature.update(getattr(ext, 'provides', []))  # Enable those flags.
			for signal in getattr(ext, 'signals', []): add_signal(signal)  # And those callbacks.
		
		# Prepare the callback cache.
		
		for ext in all:
			for signal in signals:  # Attach any callbacks that might exist.
				handler = getattr(ext, signal, None)
				if handler: signals[signal].append(handler)
			
			if hasattr(ext, '__call__'):  # This one is aliased; the extension itself is treated as WSGI middleware.
				signals['middleware'].append(ext)
		
		# Certain operations act as a stack, i.e. "before" are executed in dependency order, but "after" are executed
		# in reverse dependency order.  This is also the case with "mutate" (incoming) and "transform" (outgoing).
		for signal in inverse:
			signals[signal].reverse()
		
		# Transform the signal lists into tuples to compact them.
		self.signal = Context(**{k: tuple(v) for k, v in items(signals)})
		
		# This will save a chain() call on each request by pre-prepending the two lists.
		# Attempts to add extensions during runtime are complicated by this optimization.
		self.signal['pre'] = tuple(signals['prepare'] + signals['before'])
		
		# Continue up the chain with the `ExtensionManager` initializer, using the `web.extension` namespace.
		super(WebExtensions, self).__init__('web.extension')




# encoding: utf-8

"""Web-based REPL shell and interactive debugger extension."""

# ## Imports

from __future__ import unicode_literals

from webob.exc import HTTPNotFound
from backlash import DebuggedApplication


# ## Module Globals

log = __import__('logging').getLogger(__name__)


# ## Controller Endpoint Utility

class Console(object):
	"""Attach a console to your web application at an arbitrary location."""
	
	__slots__ = ('debugger', 'request')
	
	def __init__(self, context):
		self.debugger = context.get('debugger', None)
		self.request = context.request
	
	def __call__(self, *args, **kw):
		if not self.debugger:
			raise HTTPNotFound()
		
		return self.debugger.display_console(self.request)


# ## Extension

class DebugExtension(object):
	"""Enable an interactive exception debugger and interactive console.
	
	Possible configuration includes:
	
		* `path` -- the path to the interactive console, defaults to: `/__console__`
		* `verbose` -- show ordinarily hidden stack frames, defaults to: `False`
	"""
	
	__slots__ = ('path', 'verbose')
	
	provides = ['debugger', 'console']
	
	def __init__(self, path="/__console__", verbose=False):
		if __debug__:
			log.debug("Initializing debugger extension.")
		
		self.path = path
		self.verbose = verbose
		
		super(DebugExtension, self).__init__()
	
	def init_console(self):
		"""Add variables to the console context."""
		return dict()
	
	def init_debugger(self, environ):
		"""Add variables to the debugger context."""
		return dict(context=environ.get('context'))
	
	def __call__(self, context, app):
		"""Executed to wrap the application in middleware.
		
		The first argument is the application context, not request context.
		
		Accepts a WSGI application as the second argument and must likewise return a WSGI app.
		"""
		
		if __debug__:
			log.debug("Wrapping application in debugger middleware.")
		
		app = DebuggedApplication(
				app,
				evalex = __debug__,  # In production mode, this is a security no-no.
				show_hidden_frames = self.verbose,
				console_init_func = self.init_console,
				context_injectors = [self.init_debugger],
			)
		
		context.debugger = app
		
		return app



#!/usr/bin/env python
# encoding: utf-8

from __future__ import print_function

import os
import sys
import codecs


try:
	from setuptools.core import setup, find_packages
except ImportError:
	from setuptools import setup, find_packages


if sys.version_info < (2, 7):
	raise SystemExit("Python 2.7 or later is required.")
elif sys.version_info > (3, 0) and sys.version_info < (3, 2):
	raise SystemExit("CPython 3.3 or Pypy 3 (3.2) or later is required.")

version = description = url = author = author_email = ""  # Silence linter warnings.
exec(open(os.path.join("web", "core", "release.py")).read())  # Actually populate those values.

here = os.path.abspath(os.path.dirname(__file__))

tests_require = [
		'pytest',  # test collector and extensible runner
		'pytest-cov',  # coverage reporting
		'pytest-flakes',  # syntax validation
		'pytest-capturelog',  # log capture
		'web.dispatch.object',  # dispatch tests
		'backlash',  # debug tests
	]


# ## Package Metadata

setup(
	# ### Basic Metadata
	name = "WebCore",
	version = version,
	description = description,
	long_description = codecs.open(os.path.join(here, 'README.rst'), 'r', 'utf8').read(),
	url = url,
	download_url = 'https://github.com/marrow/WebCore/releases',
	author = author.name,
	author_email = author.email,
	license = 'MIT',
	keywords = ['marrow', 'web.core', 'web.ext'],
	classifiers = [
			"Development Status :: 5 - Production/Stable",
			"Environment :: Console",
			"Environment :: Web Environment",
			"Intended Audience :: Developers",
			"License :: OSI Approved :: MIT License",
			"Operating System :: OS Independent",
			"Programming Language :: Python",
			"Programming Language :: Python :: 2",
			"Programming Language :: Python :: 2.7",
			"Programming Language :: Python :: 3",
			"Programming Language :: Python :: 3.2",
			"Programming Language :: Python :: 3.3",
			"Programming Language :: Python :: 3.4",
			"Programming Language :: Python :: 3.5",
			"Programming Language :: Python :: Implementation :: CPython",
			"Programming Language :: Python :: Implementation :: PyPy",
			"Topic :: Internet :: WWW/HTTP :: Dynamic Content",
			"Topic :: Internet :: WWW/HTTP :: WSGI",
			"Topic :: Software Development :: Libraries",
			"Topic :: Software Development :: Libraries :: Application Frameworks",
			"Topic :: Software Development :: Libraries :: Python Modules",
		],
	# As yet unsupported by setuptools.
	#document_names = {
	#		"description": "README.rst",
	#		"license": "LICENSE.txt",
	#	},
	#contacts = [
	#		{"name": "Alice Bevan-McGregor", "email": "alice@gothcandy.com", "role": "author"},
	#	],
	#project_urls = {
	#		"Documentation": "http://pythonhosted.org/WebCore/",
	#		"Home": "https://github.com/marrow/WebCore/",
	#		"Repository": "https://github.com/marrow/WebCore/",
	#		"Tracker": "https://github.com/marrow/WebCore/issues",
	#	},
	#environments = [
	#		"python_version >= '3.2'",
	#		"'3.0' > python_version >= '2.7'",
	#	],
	
	# ### Code Discovery
	
	packages = find_packages(exclude=['bench', 'docs', 'example', 'test', 'htmlcov']),
	include_package_data = True,
	namespace_packages = [
			'web',  # primary namespace
			'web.app',  # application code goes here
			'web.ext',  # framework extensions
			'web.server',  # front-end WSGI bridges
		],
	zip_safe = True,
	
	# ### Plugin Registration
	
	entry_points = {
			# #### Re-usable applications or application components.
			'web.app': [
					'static = web.app.static:static',
				],
			
			'web.acl.predicate': [
					'not = web.ext.acl:Not',
					'always = web.ext.acl:always',
					'never = web.ext.acl:never',
					'first = web.ext.acl:First',
					'all = web.ext.acl:all',
					'any = web.ext.acl:any',
					'matches = web.ext.acl:ContextMatch',
					'contains = web.ext.acl:ContextContains',
				],
			
			# #### WebCore Extensions
			'web.extension': [
					# ##### BaseExtension, providing request, response, and default views.
					'base = web.ext.base:BaseExtension',
					'request = web.ext.base:BaseExtension',
					'response = web.ext.base:BaseExtension',
					
					# ##### Miscelaneous Builtin Extensions
					'acl = web.ext.acl:ACLExtension',  # Access control list validation.
					'analytics = web.ext.analytics:AnalyticsExtension',
					'annotation = web.ext.annotation:AnnotationExtension',  # Preferred use/needs reference.
					'cast = web.ext.annotation:AnnotationExtension',  # Legacy reference.
					'db = web.ext.db:DBExtension',  # Database Connection Management
					'typecast = web.ext.annotation:AnnotationExtension',  # Legacy reference.
					'local = web.ext.local:ThreadLocalExtension',  # Preferred use/needs reference.
					'threadlocal = web.ext.local:ThreadLocalExtension',  # Legacy reference.
					'assets = web.ext.assets:WebAssetsExtension',  # WebAssets integration.
				],
			
			# #### WSGI Server Adapters
			'web.server': [
					# These are built-in to Python itself, but are reference implementations. Do not rely on these in
					# production environments; a warning will be issued on attempts to use these if optimizations are
					# enabled.
					'wsgiref = web.server.stdlib:simple',
					'cgiref = web.server.stdlib:cgi',
					'iiscgiref = web.server.stdlib:iiscgi',  # Python 3 only!
					
					# These have additional third-party dependencies.
					# For even more, see the WSGI reference site listing of servers:
					# http://wsgi.readthedocs.org/en/latest/servers.html)
					'waitress = web.server.waitress_:serve[waitress]',  # http://s.webcore.io/aIou
					'tornado = web.server.tornado_:serve[tornado]',  # http://s.webcore.io/aIaN
					'fcgi = web.server.fcgi:serve[flup]',  # http://s.webcore.io/fhVY
					'cherrypy = web.server.cherrypy_:serve[cherrypy]',  # http://s.webcore.io/aIoF
					'appengine = web.server.appengine:serve',  # http://s.webcore.io/aIic
					'paste = paste.httpserver:serve[paste]',  # http://s.webcore.io/aIdT
					'eventlet = web.server.eventlet_:serve[eventlet]', # http://s.webcore.io/aIaa
					'gevent = web.server.gevent_:serve[gevent]',  # http://s.webcore.io/aIpU
					'diesel = web.server.diesel_:serve[diesel]',  # http://s.webcore.io/aIg2
					'bjoern = web.server.bjoern_:serve[bjoern]',  # http://s.webcore.io/aIne
				],
			
			'web.serialize': [
					'json = web.ext.serialize:json.dumps',  # JavaScript Object Notation
					'application/json = web.ext.serialize:json.dumps',  # JavaScript Object Notation
					'yaml = yaml:dumps[yaml]',  # Yet Another Markup Language
					'application/x-yaml = yaml:dumps[yaml]',  # Yet Another Markup Language
				]
		},
	
	# ## Installation Dependencies
	
	setup_requires = [
			'pytest-runner',
		] if {'pytest', 'test', 'ptr'}.intersection(sys.argv) else [],
	install_requires = [
			'marrow.package<2.0',  # dynamic execution and plugin management
			'WebOb',  # HTTP request and response objects, and HTTP status code exceptions
			'pathlib2; python_version < "3.4"',  # Path manipulation utility lib; builtin in 3.4 and 3.5.
		],
	tests_require = tests_require,
	
	extras_require = {
			# ### Recommended Environments
			'development': tests_require + [  # An extended set of useful development tools.
					'ptpython',  # Improved Python shell.  Run as "ptipython".
					'ipython',  # Additional extras to improve the Python shell.
					'pudb',  # Curses-based interactive debugger.
					'backlash',  # Web-based interactive REPL shell and traceback explorer.
					'waitress',  # Recommended development server.
				],
			'production': [  # A default set of production tools.
					'flup6',  # Python 2 and 3 compatible Flup fork.
				],
			
			# ### Dispatch Mechanisms
			'object': ['web.dispatch.object'],
			'route': ['web.dispatch.route'],
			'traversal': ['web.dispatch.traversal'],
			
			# ### General Extras
			'cli': ['web.command'],  # Command-line interface.
			'template': ['web.template', 'cinje'],  # Recommended template engine.
			'database': ['web.db', 'pymongo'],  # Recommended database engine.
			'asset': ['webassets'],  # Recommended static asset management.
			'bson': ['pymongo'],
			'yaml': ['pyyaml'],
			
			# ### Plugin Dependencies
			'waitress': ['waitress'],
			'tornado': ['tornado'],
			'flup': ['flup6'],
			'cherrypy': ['cherrypy'],
			'paste': ['paste'],
			'eventlet': ['eventlet'],
			'gevent': ['gevent'],
			'diesel': ['diesel'],
			'bjoern': ['bjoern'],
		},
)




# encoding: utf-8

"""Basic static file delivery mechanism."""

# ## Imports

from __future__ import unicode_literals

from os.path import abspath, normpath, exists, isfile, join as pathjoin, basename
from webob.exc import HTTPForbidden, HTTPNotFound


# ## Module Globals

# A standard logging object.
log = __import__('logging').getLogger(__name__)


# ## Static File Endpoint

def static(base, mapping=None, far=('js', 'css', 'gif', 'jpg', 'jpeg', 'png', 'ttf', 'woff')):
	"""Serve files from disk.
	
	This utility endpoint factory is meant primarily for use in development environments; in production environments
	it is better (more efficient, secure, etc.) to serve your static content using a front end load balancer such as
	Nginx.
	
	The first argument, `base`, represents the base path to serve files from. Paths below the attachment point for
	the generated endpoint will combine this base path with the remaining path elements to determine the file to
	serve.
	
	The second argument is an optional dictionary mapping filename extensions to template engines, for cooperation
	with the TemplateExtension.  (See: https://github.com/marrow/template)  The result of attempting to serve a
	mapped path is a 2-tuple of `("{mapping}:{path}", dict())`. For example, to render all `.html` files as Mako
	templates, you would attach something like the following:
	
		class Root:
			page = static('/path/to/static/pages', dict(html='mako'))
	
	By default the "usual culprits" are served with far-futures cache expiry headers. If you wish to change the
	extensions searched just assign a new `far` iterable.  To disable, assign any falsy value.
	"""
	
	base = abspath(base)
	
	@staticmethod
	def static_handler(context, *parts, **kw):
		path = normpath(pathjoin(base, *parts))
		
		if __debug__:
			log.debug("Attempting to serve static file.", extra=dict(
					request = id(context),
					base = base,
					path = path
				))
		
		if not path.startswith(base):  # Ensure we only serve files from the allowed path.
			raise HTTPForbidden("Cowardly refusing to violate base path policy." if __debug__ else None)
		
		if not exists(path):  # Do the right thing if the file doesn't actually exist.
			raise HTTPNotFound()
		
		if not isfile(path):  # Only serve normal files; no UNIX domain sockets, FIFOs, etc., etc.
			raise HTTPForbidden("Cowardly refusing to open a non-file." if __debug__ else None)
		
		if far and path.rpartition('.')[2] in far:
			context.response.cache_expires = 60*60*24*365
		
		if mapping:  # Handle the mapping of filename extensions to 2-tuples. 'Cause why not?
			_, _, extension = basename(path).partition('.')
			if extension in mapping:
				return mapping[extension] + ':' + path, dict()
		
		return open(path, 'rb')
	
	return static_handler




# encoding: utf-8

"""WebCore common utilities."""

# ## Imports

from __future__ import unicode_literals

from threading import RLock

from marrow.package.canonical import name


# ## Module Global Constants

sentinel = object()  # A singleton value to allow `None` as a legal value.


# ## Utility Functions

def safe_name(thing):
	"""Attempt to resolve the canonical name for an object, falling back on the `repr()` if unable to do so."""
	try:
		return name(thing)
	except:
		return repr(thing)


# ## Context-Related Utility Classes

class lazy(object):
	"""Lazily record the result of evaluating a function and cache the result.
	
	This is a non-data descriptor which tells Python to allow the instance __dict__ to override. Intended to be used
	by extensions to add zero-overhead (if un-accessed) values to the context.
	"""
	
	def __init__(self, func, name=None, doc=None):
		self.__name__ = name or func.__name__
		self.__module__ = func.__module__
		self.__doc__ = func.__doc__
		self.lock = RLock()
		self.func = func
	
	def __get__(self, instance, type=None):
		if instance is None:  # Allow direct access to the non-data descriptor via the class.
			return self
		
		with self.lock:  # Try to avoid situations with parallel thread access hammering the function.
			value = instance.__dict__.get(self.__name__, sentinel)
			
			if value is sentinel:
				value = instance.__dict__[self.__name__] = self.func(instance)
		
		return value

